"""An ox_herd plugin for posting messages to github.

This module provides a single class: PostToGitHub to post messages to github.
It is useful as a tool for the pylint plugin and perhaps others.

Provided that you include this plugin in the OX_PLUGINS variable in
ox_herd.settings or in your OX_PLUGINS environment variable, this will be
picked up automatically and used as a plugin.

In general, there are a variety of more complicated things you can do
in configuring plugins. See documentation on plugins or see the pytest_plugin
example for more details.
"""

import configparser
import json

from eyap.core import github_comments

from ox_herd.core.plugins import base
from ox_herd.core.ox_tasks import OxHerdTask


class PostToGitHub(OxHerdTask, base.OxPluginComponent):
    """Class to post a message to github.

    """

    def __init__(self, msg, full_repo, title, number, conf_file, conf_sec,
                 *args, **kw):
        """Initializer.

        :arg msg:      String message to post.

        :arg full_repo: Full name of github repo (e.g., 'aocks/ox_herd').

        :arg title:     String title of issue to post to *ONLY* if no
                        github_issue is specified in the conf file (see below).
                        This will be ignored if github_issue is provided.

        :arg number:    Optional issue number. This is useful if you have
                        multiple issues with the same title and need to
                        distinguish between them.

        :arg conf_file: Path to configuration file to be read by python
                        configparser module.

        :arg conf_sec:  String name of section in conf_file to read. 
                        This section should have entries for the following:

                           github_user:  Name of github user for login.
                           github_token: Token or password to access github.
                           github_issue: Optional issue title to use instead of
                                         title argument.
        
        :arg *args:    Argumnets to OxHerdTask.__init__.
        
        :arg **kw:     Keyword arguments to OxHerdTask.__init__.
        
        """
        OxHerdTask.__init__(self, *args, **kw)
        base.OxPluginComponent.__init__(self)
        self.msg = msg
        self.full_repo = full_repo
        self.title = title
        self.number = number
        self.conf_file = conf_file
        self.conf_sec = conf_sec

    @classmethod
    def main_call(cls, ox_herd_task):
        """Main method to post to github.

        :arg ox_herd_task:   Instance of a PostToGitHub task perhaps containing
                             additional data (e.g., ox_herd_task.name). 

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:       Dictionary with 'return_value' and 'json_blob' as
                        required for OxPluginComponent.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:        Post a message to github.

        """
        my_config = configparser.ConfigParser()
        my_config.read(ox_herd_task.conf_file)
        my_csec = my_config[ox_herd_task.conf_sec]
        cthread = cls.prep_comment_thread(
            ox_herd_task.title, ox_herd_task.number, ox_herd_task.full_repo,
            my_csec)
        cthread.add_comment(ox_herd_task.msg, allow_create=True)

        return {
            'return_value': 'Task %s completed succesfully.' % (
                ox_herd_task.name), 'json_blob': json.dumps({})}


    @staticmethod
    def prep_comment_thread(title, number, full_repo, my_conf):
        """Prepare a CommentThread object to use in positing comments.
        
        :arg ox_herd_task: Ox Herd task with raw data.       
        
        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
        
        :returns:  A GitHubCommentThread object.
        
        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
        
        PURPOSE:   This method reads from the configuration in the dictionary
                   in my_conf, figures out the github parameters, and creates a
                   GitHubCommentThread we can use in posting comments.
        """
        user = my_conf['github_user']
        token = my_conf['github_token']
        topic = my_conf['github_issue'] if 'github_issue' in my_conf else None

        if topic is None:
            topic = title
            thread_id = number
            if isinstance(thread_id, str):
                thread_id = thread_id.strip()
            if thread_id == '':
                thread_id = None
        else:
            thread_id = None

        owner, repo = full_repo.split('/')

        comment_thread = github_comments.GitHubCommentThread(
            owner, repo, topic, user, token, thread_id=thread_id)

        return comment_thread        



"""Module for scheduling jobs.
"""

import logging

try: # try to import rq_scheduler and redis but allow other modes if fail
    import rq
    from rq.job import Job, UnpickleError
    from rq import get_failed_queue, Queue
    import rq_scheduler
    from redis import Redis
except Exception as problem:
    logging.error('Could not import rq_scheduler and redis because %s.\n%s',
                  str(problem), 'Continue with non-rq options.')
    

class OxScheduler(object):

    @classmethod
    def add_to_schedule(cls, ox_herd_task, manager):
        name = 'schedule_via_%s' % manager
        sched_func = getattr(cls, name)
        return sched_func(ox_herd_task)

    @staticmethod
    def schedule_via_instant(ox_herd_task):
        return ox_herd_task.func(ox_herd_task=ox_herd_task)
        
    @staticmethod
    def schedule_via_rq(ox_herd_task):
        rq_kw = dict([(name, getattr(ox_herd_task, name)) 
                      for name in ox_herd_task.rq_fields])
        scheduler = rq_scheduler.Scheduler(
            connection=Redis(), queue_name=rq_kw.pop('queue_name'))
        if rq_kw['cron_string']:
            return scheduler.cron(
                rq_kw.pop('cron_string'), rq_kw.pop('func'),
                kwargs={'ox_herd_task' : ox_herd_task}, **rq_kw)
        else:
            raise ValueError('No scheduling method for rq task.')

    @staticmethod
    def cancel_job(job):
        scheduler = rq_scheduler.Scheduler(connection=Redis())
        return scheduler.cancel(job)

    @staticmethod
    def cleanup_job(job_id):
        conn = Redis()
        failed_queue = get_failed_queue(conn)
        failed_queue.remove(job_id)
        return 'Removed job %s' % str(job_id)


    @staticmethod
    def requeue_job(job_id):
        conn = Redis()
        failed_queue = get_failed_queue(conn)
        result = failed_queue.requeue(job_id)
        return result

    @classmethod
    def launch_job(cls, job_id):
        logging.warning('Preparing to launch job with id %s', str(job_id))
        old_job = cls.find_job(job_id)
        old_args = old_job.kwargs['ox_herd_task']
        my_args = old_args.make_copy(old_args)
        return cls.launch_raw_task(my_args)

    @staticmethod
    def launch_raw_task(raw_task):
        """Launch an OxHerdTask instance into python rq.
        
        :arg raw_task:    OxHerdTask instance to run.
        
        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
        
        :returns:  Newly launched job.
        
        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
        
        PURPOSE:   Make it easy to launch an OxHerdTask into python rq
                   right now instead of waiting for cron or scheduler.
        
        """
        rq_kw = dict([(name, getattr(raw_task, name)) 
                      for name in raw_task.rq_fields])
        if 'cron_string' in rq_kw:
            rq_kw.pop('cron_string')  # launching now so get rid of cron_string

        queue_name = rq_kw.pop('queue_name')
        my_queue = rq.Queue(queue_name, connection=Redis())
        my_func = rq_kw.pop('func')
        new_job = my_queue.enqueue(
            my_func, kwargs={'ox_herd_task' : raw_task}, **rq_kw)
        logging.info('Launched job on q=%s with kwargs=%s; rq_kw=%s',
                     queue_name, raw_task, rq_kw)

        return new_job

        

    @staticmethod
    def find_job(target_job):
        scheduler = rq_scheduler.Scheduler(connection=Redis())
        job = Job.fetch(target_job, connection=scheduler.connection)
        if job:
            return job

        # FIXME: stuff below probably obsolete

        job_list = scheduler.get_jobs()
        for job in job_list:
            if job.id == target_job:
                return job
        return None

    @staticmethod
    def get_failed_jobs():
        results = []
        conn = Redis()
        failed = get_failed_queue(conn)
        failed_jobs = failed.jobs
        for item in failed_jobs:
            try:
                kwargs = getattr(item, 'kwargs', {})
            except UnpickleError as problem:
                logging.info('Could not unickle %s because %s; skip',
                             str(item), str(problem))
                continue
            if kwargs.get('ox_herd_task', None) is not None:
                results.append(item)
        return results

    @staticmethod
    def get_scheduled_jobs():
        results = []
        scheduler = rq_scheduler.Scheduler(connection=Redis())
        jobs = scheduler.get_jobs()
        for item in jobs:
            try:
                if not getattr(item, 'kwargs').get('ox_herd_task', None):
                    continue
            except UnpickleError as problem:
                logging.info('Could not unickle %s because %s; skip',
                             str(item), str(problem))
                continue
            try:
                cron_string = item.meta.get('cron_string', None)
                if cron_string:
                    results.append(item)
                else:
                    logging.info('Skipping task without cron_string.'
                                 'Probably was just a one-off launch.')
            except Exception as problem:
                logging.warning(
                    'Skip job %s in get_scheduled_jobs due to exception %s',
                    str(item), problem)

        return results

    @staticmethod
    def get_queued_jobs(allowed_queues=None):
        queue = Queue(connection=Redis())
        all_jobs = queue.jobs
        if not allowed_queues:
            return all_jobs
        else:
            return [j for j in all_jobs if j.origin in allowed_queues]


    @classmethod
    def add_task_if_unscheduled(cls, task_list, manager='rq'):
        scheduled_jobs = cls.get_scheduled_jobs()
        sj_dict = {}
        for item in scheduled_jobs:
            task = item.kwargs['ox_herd_task']
            if task.name in sj_dict:
                logging.warning('Task %s seen multiple times.', task.name)
            sj_dict[task.name] = task
        for task in task_list:
            if task.name in sj_dict:
                logging.info('Not scheduling %s since already scheduled.',
                             task.name)
            else:
                logging.info('Adding %s to schedule.', task.name)
                cls.add_to_schedule(task, manager)



"""Script to schedule execution of tests automatically.
"""

import logging
import collections
import argparse

from ox_herd.core.scheduling import OxScheduler

def make_manager_choices():
    "Return dictionary of manager choices and docs."

    return collections.OrderedDict([
        ('rq', 'python-rq backend for automated background runs'),
        ('instant', 'Run instantly (useful for testing).'),
    ])

def prepare_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    managers = make_manager_choices()
    parser.add_argument('--url', required=True, help=(
        'Location for package to test. This can be a github URL or a file\n'
        'location.'))
    parser.add_argument('--name', default='test_', help=(
        'String name for the job you are going to schedule.'))
    parser.add_argument('--manager', required=1, choices=list(managers), help=(
        'Backend manager for automated testing. Options include:\n  '
        + '\n  '.join([
            '%s : %s' % (name, docs) for name, docs in managers.items()])))
    parser.add_argument('--json_file', help=(
        'Optional path to json file for output of test run. Usually,\n'
        'you should not provide this and a temp file will be used\n'
        'and then deleted. If you do manually specify a path, then that\n'
        'path will be used and *NOT* deleted (which can be useful\n'
        'for debugging).'))
    parser.add_argument('--timeout', type=int, default=900, help=(
        'Timeout in seconds to allow for tasks.'))
    parser.add_argument('--cron_string', help=(
        'Cron format string for when to schedule the task.\n'
        'For example, "5 1 * * 3" would be every Wednesday at 1:05 am.\n'
        'This is used for --manager choices such as rq which support cron\n'
        'scheduling. NOTE: cron_string should have 5 fields. If you try to\n'
        'use the non-standard extended cron format with 6 fields, you may get\n'
        'unexpected results.'))
    parser.add_argument('-p', '--pytest', action='append', help=(
        'Provide the given argument to pytest. For example, if you did\n'
        '-p=--ignore -p=/path/to/setup.py then you could pass an ignore\n'
        'flag to pytest.'))

    return parser

def run():
    parser = prepare_parser()
    args = parser.parse_args()
    sched = OxScheduler.add_to_schedule(args)

if __name__ == '__main__':
    run()



"""Module to represent basic settings for ox_herd package.
"""

import os

# Path to your ox herd configuration file.
OX_HERD_CONF = os.environ.get('OX_HERD_CONF', os.path.join(
    os.environ.get('HOME', ''), '.ox_herd_conf'))

# Stub user database used only for testing stand alone ox_herd.
# You can either manually configure this to have keys as usernames
# and password hashes created according to the following:
#
# >>> from passlib.apps import custom_app_context as pwd_context
# >>> hash = pwd_context.hash("somepass")
#
# or provide a [STUB_USER_DB] section in OX_HERD_CONF with username
# as keys and password hashes as values
#
#                       WARNING!
#
# If you really want security, you will also need to setup SSL.
# You should probably setup SSL and some better password security
# in your app and just use ox_herd as a Flask blueprint to inherit
# proper security instead of using the login stub but it is there for
# ease of testing and local use.
STUB_USER_DB = {}

# Space separated list of queue names to show in showing scheduled tests.
# Can be changed by your own software after importing settings if you like.
# First element will be the default queue to use.
QUEUE_NAMES = 'default'

# Optional pair representing mode and string path to where we store
# database tracking job execution. Default mode is ('redis', None)
# to just use redis. You can also use 'sqlite' with
# either a path to the sqliet db or None to use a default path.
RUN_DB = ('redis', None)

# Prefix to use in things we store on redis queue for ox_herd.
REDIS_PREFIX = 'ox_herd:'

# List of names of plugins to enable. Each "name" is a string which
# can be used in a python import statement. By default we enable the
# pytest_plugin provided with ox_herd as an example.
# See ox_herd/core/plugins.py for description of how plugins work.
OX_PLUGINS = [
    'ox_herd.core.plugins.pytest_plugin',
    'ox_herd.core.plugins.pylint_plugin',
    'ox_herd.core.plugins.example_psutil_plugin',
    ]



"""Forms for ox_herd commands.
"""

from wtforms import StringField
from ox_herd.core.plugins import base

class SchedJobForm(base.GenericOxForm):
    """Use this form to enter parameters for a new pytest job to schedule.
    """

    url = StringField('url', [], description=(
        'Location for package to test. This can be a github URL or a file\n'
        'location.'))

    pytest_cmd = StringField(
        'pytest_cmd', [], 
        default='--ignore=/path/to/setup.py --doctest-modules',
        description=(
            'Command line arguments for pytest. For example, you could\n'
            'provide something like\n\n'
            '  --ignore=/path/to/setup.py --doctest-modules\n\n'
            'to get pytest working as you like. This will be passed through\n'
            'shlex.split.\n'))
    
    xml_file = StringField(
        'xml_file', [], description=(
            'Optional path to xml file for output of test run. Usually,\n'
            'you should not provide this and a temp file will be used\n'
            'and then deleted. If you do manually specify a path, then that\n'
            'path will be used and *NOT* deleted (which can be useful\n'
            'for debugging).'))



"""Module containing some plugin to run pylint.
"""

import re
import datetime
import io
import logging
import os
import urllib
import urllib.parse
import urllib.request

from pylint import lint
from pylint.reporters.text import ParseableTextReporter

from ox_herd.core.plugins import base
from ox_herd.core.ox_tasks import OxHerdTask
from ox_herd.core.plugins.pylint_plugin import forms


class OxHerdPyLintPlugin(base.OxPlugin):
    """Plugin to provide pylint services for ox_herd
    """

    __blueprint = None

    @classmethod
    def set_bp(cls, my_bp):
        cls.__blueprint = my_bp
    
    def get_flask_blueprint(self):
        """Implement as required by OxPlugin."""

        return self.__class__.get_bp()

    @classmethod
    def get_bp(cls):
        return cls.__blueprint
    

    def name(self):
        """Implement as required by OxPlugin."""

        return 'pylint_plugin'

    def description(self):
        """Implement as required by OxPlugin."""

        return "Plugin to provide pylint services for ox_herd."

    def get_components(self):
        return [RunPyLint('plugin component')]

def make_kill_regexps():
    """Make dictionary identifying regular expressions to remove from output.

    Some of the checkers provide results that confuse pylint. So we remove
    those.
    """
    result = {
        'long_dash': r'^-+ *',
        'pylint_rating' : '^[Yy]our code has been rated.*$'
    }
    return result


class RunPyLint(OxHerdTask, base.OxPluginComponent):
    """Run pylint to analyze code quality
    """

    def __init__(self, *args, url=None, **kw):
        """Initializer.
        
        :arg *args:    Argumnets to OxHerdTask.__init__.

        :arg url:      URL representing where to run pylint on.  
        
        :arg **kw:     Keyword arguments to OxHerdTask.__init__.
        
        """
        OxHerdTask.__init__(self, *args, **kw)
        self.url = url

    @staticmethod
    def cmd_name():
        """Provide name as required by OxPluginComponent.
        """
        return 'pylint'

    @staticmethod
    def get_template_name():
        "Override to use custom template to report py lint results."

        return 'py_lint_report.html'

    @staticmethod
    def run_pylint(filename):
        """Run pylint on the given file.
        """
        args = ["-r","n"]
        pylint_output = io.StringIO()
        lint.Run([filename]+args, 
                 reporter=ParseableTextReporter(pylint_output), exit=False)
        pylint_output.seek(0)
        result = pylint_output.read()
        kill_regexps = make_kill_regexps()
        for re_name, my_re in kill_regexps.items():
            logging.debug('Cleaning output with re %s', re_name)
            result = re.sub(my_re, '', result, flags=re.M)
        return result

    @classmethod
    def main_call(cls, ox_herd_task):
        results = cls.do_lint(ox_herd_task)
        rval = 'completed lint: ' + ', '.join(['%s=%s' % (
            name, results['summary'][name]) for name in [
                'passed', 'issues', 'failures']])
        return {'return_value' : rval, 'json_blob' : results}

    @staticmethod
    def lint_results_to_dict(lint_output):
        """Take string output of lint command and return dictionary summary.
        """
        data = [line for line in lint_output.split('\n') if (
            line and line.strip() and line[:10] != ('*'*10))]
        if not data:
            return {'outcome' : 'success'}
        result = {'outcome' : 'issues', 'count' : len(data), 
                  'issues' : [line.split(':') for line in data]}
        return result

    @classmethod
    def do_lint(cls, py_lint_args):
        """Do pylint on desired URL.
        
        :arg py_lint_args:   Instance of class with arguments.
        
        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
        
        :returns:       A list of dicts intended to be serialized into
                        JSON format outlining the results of the pylint run.
        
        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
        
        PURPOSE:        Analyze code quality.
        
        """
        results = {'lints' : [], 'created_at' : str(datetime.datetime.utcnow()),
                   'task_name' : py_lint_args.name, 'url' : py_lint_args.url}
        url = urllib.parse.urlparse(py_lint_args.url)
        passed = 0
        issues = 0
        failures = 0
        if url.scheme == 'file':
            for root, dummy_dirs, files in os.walk(url.path):
                for my_file in files:
                    if my_file[-3:] == '.py':
                        my_file = os.path.join(root, my_file)
                        logging.info('Checking %s.', my_file)
                        my_result = {'file' : re.sub(
                            '^%s' % url.path, '', my_file).lstrip('/')}
                        try:
                            lint_results = cls.run_pylint(my_file)
                            my_result.update(cls.lint_results_to_dict(
                                lint_results))
                            if my_result['outcome'] == 'success':
                                passed += 1
                            else:
                                assert my_result['outcome'] == 'issues'
                                issues += 1
                        except Exception as bad: #pylint: disable=broad-except
                            failures += 1
                            my_result['outcome'] = 'failed'
                            my_result['longrepr'] = str(bad)
                        results['lints'].append(my_result)
            results['summary'] = {
                'issues' : issues, 'passed' : passed, 'failures' : failures}
            return results
        else:
            raise ValueError('URL scheme of %s not handled yet.' % url.scheme)

    def get_ox_task_cls(self):
        return self.__class__

    def get_flask_form(self):
        return forms.PylintForm




"""Views for ox_herd flask blueprint.
"""

import copy
import logging
import os
import collections

import markdown

from flask import render_template, redirect, request, Markup, url_for
from flask_login import login_required

from ox_herd.ui.flask_web_ui.ox_herd import OX_HERD_BP
from ox_herd.core import scheduling, simple_ox_tasks, ox_run_db
from ox_herd import settings
from ox_herd.core.plugins import manager as plugin_manager

from collections import namedtuple
def d_to_nt(dictionary):
    "Convert dictionaryt to namedtuple"
    return namedtuple('GenericDict', dictionary.keys())(**dictionary)

def delete_old_data(old_data):
    for old_time, old_file in old_data:
        logging.debug('Deleting old file %s.', old_file)
        os.remove(old_file)

@OX_HERD_BP.route('/')
@OX_HERD_BP.route('/index')
@OX_HERD_BP.route('/show_index')
@login_required
def index():
    """Main page for ox_herd.

    Usually will not be true root since blueprint will
    be registered with url_prefix.
    """
    commands = collections.OrderedDict([
        (name, Markup('<A HREF="%s">%s</A>' % (
            url_for('ox_herd.%s' % name), name))) for name in [
                'show_plugins', 'list_tasks', 'show_scheduled', 'show_task_log',
                'cancel_job', 'cleanup_job']])

    return render_template('ox_herd/templates/intro.html', commands=commands)


@OX_HERD_BP.route('/list_tasks')
@login_required
def list_tasks():
    "Show list of tasks so you can inspect them."

    tasks = ox_run_db.create().get_tasks()
    return render_template('task_list.html', title='Task List', tasks=tasks)


@OX_HERD_BP.route('/show_task_log')
@login_required
def show_task_log():
    "Show log of tasks run."
    
    run_db = ox_run_db.create()
    start_utc=request.args.get('start_utc', None)
    end_utc=request.args.get('end_utc', None)
    tasks = run_db.get_tasks(start_utc=start_utc, end_utc=end_utc)
    other = []
    task_dict = collections.OrderedDict([('started', []), ('finished', [])])
    for item in reversed(sorted(
            tasks, key=lambda t: (t.task_end_utc, t.task_start_utc))):
        if item.task_status in task_dict:
            task_dict[item.task_status].append(item)
        else:
            other.append(item)
    task_dict['other'] = other
    
    return render_template(
        'task_log.html', title='Task log', start_utc=start_utc,
        end_utc=end_utc, task_dict=task_dict)


@OX_HERD_BP.route('/show_task')
@login_required
def show_task():
    "Show information about a task."

    task_id = request.args.get('task_id', None)
    run_db = ox_run_db.create()
    try:
        task_data = run_db.get_task(task_id)
        if not task_data:
            raise KeyError('No task with id %s' % str(task_id))
    except Exception as problem:
        return render_template(
            'generic_error.html',title='Could not find task',
            commentary='Could not find task with id %s because %s' % (
                task_id, problem))

    template = request.args.get('template', task_data.template)
    template = template if (
        template and template.strip() and template != 'default') else (
            'generic_ox_task_result.html')
    return render_template(template, title='Task Report', task_data=task_data)

@OX_HERD_BP.route('/show_scheduled')
@login_required
def show_scheduled():
    queue_names = request.args.get('queue_names', settings.QUEUE_NAMES)
    queue_names = list(sorted(queue_names.split()))
    my_jobs = scheduling.OxScheduler.get_scheduled_jobs()
    failed_jobs = scheduling.OxScheduler.get_failed_jobs()
    queued = scheduling.OxScheduler.get_queued_jobs(queue_names)
    return render_template('task_schedule.html', task_schedule=my_jobs,
                           queue_names=queue_names, failed_jobs=failed_jobs,
                           queued=queued)

@OX_HERD_BP.route('/show_job')
@login_required
def show_job():
    jid = request.args.get('jid', None)
    if not jid:
        return redirect(url_for('ox_herd.index'))
    else:
        my_job = scheduling.OxScheduler.find_job(jid)
        ox_herd_task = getattr(my_job, 'kwargs', {}).get('ox_herd_task', None)
        if ox_herd_task is None:
            raise ValueError('Job for id %s has no ox_herd_task (job is %s)' % (
                jid, str(my_job))) # FIXME: should show nice error not raise
        return render_template('job_info.html', item=my_job)


@OX_HERD_BP.route('/launch_job')
@login_required
def launch_job():
    jid = request.args.get('jid', None)
    if jid:
        new_job = scheduling.OxScheduler.launch_job(jid)
        new_jid = new_job.id
    else:
        new_jid = None
    
    return render_template('launch_job.html', jid=new_jid)

@OX_HERD_BP.route('/cancel_job')
@login_required
def cancel_job():
    jid = request.args.get('jid', None)
    if not jid:
        return render_template('cancel_job.html')
    else:
        cancel = scheduling.OxScheduler.cancel_job(jid)
        return render_template('cancel_job.html', jid=jid, cancel=cancel)

@OX_HERD_BP.route('/cleanup_job')
@login_required
def cleanup_job():
    cleanup = None
    jid = request.args.get('jid', None)
    if jid:
        cleanup = scheduling.OxScheduler.cleanup_job(jid)
    return render_template('cleanup_job.html', jid=jid, cleanup=cleanup)


@OX_HERD_BP.route('/requeue_job')
@login_required
def requeue_job():
    jid = request.args.get('jid', None)
    if jid:
        requeue = scheduling.OxScheduler.requeue_job(jid)
    return render_template('requeue_job.html', jid=jid, requeue=requeue)



@OX_HERD_BP.route('/show_plugins')
@login_required
def show_plugins():
    actives = plugin_manager.PluginManager.get_active_plugins()
    components = []
    for name, plug in actives.items():
        comp_list = plug.get_components()
        logging.debug('Processing plugin %s.', name)
        urls = [(c.cmd_name(), '%s?plugname=%s&plugcomp=%s' % (
            url_for('ox_herd.use_plugin'), name, c.cmd_name()))
                for c in comp_list]
        components.append((name, urls))
    return render_template('show_plugins.html', components=components)


@OX_HERD_BP.route('/use_plugin', methods=['GET', 'POST'])
@login_required
def use_plugin():
    plugname = request.args.get('plugname', '').strip()
    plugcomp = request.args.get('plugcomp', '').strip()
    plugdict = plugin_manager.PluginManager.get_active_plugins()
    if plugname not in plugdict:
        return 'FIXME: plugin named %s not found in %s' % (plugname, list(
            plugdict))
    myplug = plugdict[plugname]
    compdict = dict([(c.cmd_name(), c) for c in myplug.get_components()])
    if plugcomp not in compdict:
        return 'FIXME: component named %s not found' % plugcomp
    my_comp = compdict[plugcomp]
    my_form_cls = my_comp.get_flask_form()
    my_form = my_form_cls()
    if my_form.validate_on_submit():
        klass = my_comp.get_ox_task_cls()
        info = klass(name=my_form.name.data)
        my_form.populate_obj(info)
        job = scheduling.OxScheduler.add_to_schedule(info, info.manager)
        return redirect('%s?jid=%s' % (url_for('ox_herd.show_job'), job.id))
    else:
        template = my_comp.get_flask_form_template()
        intro=Markup(markdown.markdown(my_form.__doc__, extensions=[
            'fenced_code', 'tables']))
        return render_template(
            template, form=my_form, intro=intro, title=(
                'Form for component %s of plugin %s' % (plugcomp, plugname)))
            

@OX_HERD_BP.route('/schedule_job', methods=['GET', 'POST'])
@login_required
def schedule_job():

    jid = request.args.get('jid', None)
    my_job = scheduling.OxScheduler.find_job(jid)
    my_args = my_job.kwargs.get('ox_herd_task', None)
    if my_args is None:
        raise ValueError("job %s had no kwargs['ox_herd_task']"%str(my_job))
    my_args = copy.deepcopy(my_args)
    my_form_cls = my_args.get_flask_form()
    my_form = my_form_cls(obj=my_args)

    if my_form.validate_on_submit():
        my_form.populate_obj(my_args)
        job = scheduling.OxScheduler.add_to_schedule(my_args, getattr(
            my_args, 'manager', 'rq'))
        return redirect('%s?jid=%s' % (url_for('ox_herd.show_job'), job.id))

    my_form.name.data += '_copy' # change name to add _copy if making new job

    return render_template(
        'ox_wtf.html', form=my_form, title='Schedule Test',
        intro=Markup(markdown.markdown(my_form.__doc__, extensions=[
            'fenced_code', 'tables'])))


@OX_HERD_BP.route('/delete_task_from_db')
@login_required
def delete_task_from_db():
    """Delete a task from the task database.

    This is mainly intended to be used from the task list and not so
    useful directly since it requires the task_id.
    """
    task_id = request.args.get('task_id', None)
    if task_id:
        run_db = ox_run_db.create()
        run_db.delete_task(task_id)
        return render_template('generic_display.html', commentary=(
            'Delete task with id %s from database.' % task_id))
    return render_template('generic_display.html', commentary=(
        'Found no task_id so cannot do anything.'))
